@@ -10,8 +10,10 @@ use crate::{
     util::{dict_to_list, filter_out_idx, is_unique, vec_map_insert, EinInt, ALPHABET},
 };
 use itertools::{izip, Itertools};
+use once_cell::sync::Lazy;
 use pyo3::exceptions;
 use pyo3::types::IntoPyDict;
+use regex::Regex;
 use std::{fmt::Debug, iter::zip};
 
 use thiserror::Error;
@@ -260,9 +262,9 @@ impl RearrangeSpec {
         result
     }
 
-    pub fn to_einops_string(&self) -> String {
+    pub fn to_einops_string(&self, size_on_letter: bool) -> String {
         let ints_in_output = self.ints_in_out();
-
+        let ints_needing_sizes = self.ints_needing_sizes();
         let input_letters = self
             .input_ints
             .iter()
@@ -271,7 +273,18 @@ impl RearrangeSpec {
                     .iter()
                     .map(|i| {
                         if ints_in_output.contains(i) || self.int_sizes[*i as usize].is_none() {
-                            ALPHABET[*i as usize].to_owned()
+                            if size_on_letter
+                                && ints_needing_sizes.contains(i)
+                                && self.int_sizes[*i as usize].is_some()
+                            {
+                                format!(
+                                    "{}:{}",
+                                    ALPHABET[*i as usize].to_owned(),
+                                    Option::<usize>::from(self.int_sizes[*i as usize]).unwrap()
+                                )
+                            } else {
+                                ALPHABET[*i as usize].to_owned()
+                            }
                         } else {
                             self.int_sizes[*i as usize].0.to_string()
                         }
@@ -294,7 +307,18 @@ impl RearrangeSpec {
                     .iter()
                     .map(|i| {
                         if ints_in_input.contains(i) || self.int_sizes[*i as usize].is_none() {
-                            ALPHABET[*i as usize].to_owned()
+                            if size_on_letter
+                                && ints_needing_sizes.contains(i)
+                                && self.int_sizes[*i as usize].is_some()
+                            {
+                                format!(
+                                    "{}:{}",
+                                    ALPHABET[*i as usize].to_owned(),
+                                    Option::<usize>::from(self.int_sizes[*i as usize]).unwrap()
+                                )
+                            } else {
+                                ALPHABET[*i as usize].to_owned()
+                            }
                         } else {
                             self.int_sizes[*i as usize].0.to_string()
                         }
@@ -311,6 +335,75 @@ impl RearrangeSpec {
         input_letters + " -> " + &output_letters
     }
 
+    #[staticmethod]
+    pub fn from_einops_string(string: &str) -> Option<Self> {
+        static RE_TOP_LEVEL: Lazy<Regex> = Lazy::new(|| {
+            Regex::new(r"(\d+)|(\w[\w\d]*(?::\d+)?)|\(\s*((?:(\w[\w\d]*(?::\d+)?|\d+)\s*)*)\)")
+                .unwrap()
+        });
+
+        if let [input, output] = string.split("->").map(str::trim).collect::<Vec<_>>()[..] {
+            let mut sizes: OpShape = sv![];
+            let mut string_to_int: HashMap<String, u8> = HashMap::new();
+            let mut f = |s: &str| -> Option<RInts> {
+                let handle_int = |num, sizes: &mut OpShape| {
+                    sizes.push(Some(num).into());
+                    (sizes.len() - 1) as u8
+                };
+                let mut handle_string_nosize = |s: String, sizes: &mut OpShape| {
+                    if let Some(old) = string_to_int.get(&s) {
+                        *old
+                    } else {
+                        sizes.push(None.into());
+                        string_to_int.insert(s.to_owned(), (sizes.len() - 1) as u8);
+                        (sizes.len() - 1) as u8
+                    }
+                };
+                let mut handle_string = |string: String, sizes: &mut OpShape| -> Option<u8> {
+                    if let [string, num] = string.split(":").collect::<Vec<_>>()[..] {
+                        let num = num.parse::<usize>().ok()?;
+                        let result = handle_string_nosize(string.to_owned(), sizes);
+                        sizes[result as usize] = Some(num).into();
+                        Some(result)
+                    } else {
+                        Some(handle_string_nosize(string, sizes))
+                    }
+                };
+                RE_TOP_LEVEL
+                    .captures_iter(s)
+                    .map(|tl| {
+                        if let Some(c) = tl.get(1) {
+                            let num = c.as_str().parse::<usize>().ok()?;
+                            Some(sv![handle_int(num, &mut sizes)])
+                        } else if let Some(c) = tl.get(2) {
+                            Some(sv![handle_string(c.as_str().to_owned(), &mut sizes)?])
+                        } else if let Some(c) = tl.get(3) {
+                            c.as_str()
+                                .split_whitespace()
+                                .filter(|z| !z.is_empty())
+                                .map(|s| {
+                                    s.parse::<usize>()
+                                        .map(|z| Some(handle_int(z, &mut sizes)))
+                                        .unwrap_or_else(|_e| {
+                                            handle_string(s.to_owned(), &mut sizes)
+                                        })
+                                })
+                                .collect::<Option<_>>()
+                        } else {
+                            None
+                        }
+                    })
+                    .collect()
+            };
+            return Some(RearrangeSpec {
+                input_ints: f(input)?,
+                output_ints: f(output)?,
+                int_sizes: sizes,
+            });
+        }
+        None
+    }
+
     pub fn add_batch_dims(&self, batch_rank: usize) -> Self {
         let to_prepend =
             || (self.int_sizes.len()..self.int_sizes.len() + batch_rank).map(|x| sv![x as u8]);
@@ -329,17 +422,18 @@ impl RearrangeSpec {
                 .collect(),
         }
     }
-
-    pub fn letter_sizes(&self) -> Vec<(String, u64)> {
-        // because to_einops_string uses numerals on all added dims, we only need letter sizes for splits
-        let ints_needing_sizes: HashSet<EinInt> = self
-            .input_ints
+    pub fn ints_needing_sizes(&self) -> HashSet<EinInt> {
+        self.input_ints
             .iter()
             .filter(|x| x.len() > 1)
             .flatten()
             .copied()
-            .collect();
-        ints_needing_sizes
+            .collect()
+    }
+    pub fn letter_sizes(&self) -> Vec<(String, u64)> {
+        // because to_einops_string uses numerals on all added dims, we only need letter sizes for splits
+
+        self.ints_needing_sizes()
             .into_iter()
             .map(|x| {
                 let size_here = self.int_sizes[x as usize];
@@ -348,9 +442,11 @@ impl RearrangeSpec {
             })
             .collect()
     }
+
     pub fn to_einops_string_and_letter_sizes(&self) -> (String, Vec<(String, u64)>) {
-        (self.to_einops_string(), self.letter_sizes())
+        (self.to_einops_string(false), self.letter_sizes())
     }
+
     /// apply rearrange to tensor
     pub fn apply(&self, tensor: &PyAny) -> PyResult<Py<PyAny>> {
         let (string, letter_sizes) = self.to_einops_string_and_letter_sizes();
