@@ -1,5 +1,6 @@
 use crate::hashmaps::FxHashMap as HashMap;
 use crate::py_types::PY_UTILS;
+use crate::rrfs::{save_tensor_rrfs, tensor_from_hash_prefix};
 use crate::smallvec::Sv;
 use crate::{
     circuit::{CircuitConstructionError, HashBytes},
@@ -371,7 +372,7 @@ pub struct TensorIndex(pub Vec<TensorAxisIndex>);
 
 impl Display for TensorIndex {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        let mut result = "(".to_owned();
+        let mut result = "[".to_owned();
         for (i, idx) in self.0.iter().enumerate() {
             result.push_str(&match idx {
                 TensorAxisIndex::Single(i) => i.to_string(),
@@ -388,7 +389,7 @@ impl Display for TensorIndex {
                 result.push(',');
             }
         }
-        result.push(')');
+        result.push(']');
         write!(f, "{}", result)
     }
 }
@@ -415,6 +416,7 @@ impl TensorIndex {
             })
             .collect()
     }
+
     pub fn is_identity(&self, shape: &Shape) -> bool {
         zip(&self.0, shape).all(|(idx, &l)| match idx {
             TensorAxisIndex::Single(_i) => false,
@@ -532,6 +534,75 @@ impl TensorIndex {
         }
         hasher.finalize().into()
     }
+
+    pub fn repr_bijection(&self) -> String {
+        let mut result = "[".to_owned();
+        for (i, idx) in self.0.iter().enumerate() {
+            result.push_str(&match idx {
+                TensorAxisIndex::Single(i) => i.to_string(),
+                TensorAxisIndex::Slice(slice) => {
+                    slice.start.map(|i| i.to_string()).unwrap_or("".to_owned())
+                        + ":"
+                        + &slice.stop.map(|i| i.to_string()).unwrap_or("".to_owned())
+                }
+                TensorAxisIndex::Tensor(tensor) => {
+                    save_tensor_rrfs(tensor.clone()).unwrap();
+                    "t".to_owned() + &base16::encode_lower(&tensor.hash().unwrap())[..10]
+                }
+            });
+            if i != self.0.len() - 1 {
+                result.push(',');
+            }
+        }
+        result.push(']');
+        result
+    }
+    pub fn from_bijection_string(string: &str) -> Result<Self, CircuitConstructionError> {
+        let er = CircuitConstructionError::ParsingFail {
+            string: string.to_owned(),
+        };
+        Ok(TensorIndex(
+            string
+                .strip_prefix("[")
+                .ok_or(er.clone())?
+                .strip_suffix("]")
+                .ok_or(er.clone())?
+                .split(",")
+                .map(str::trim)
+                .filter(|x| !x.is_empty())
+                .map(|x| {
+                    let num_er = CircuitConstructionError::ParsingNumberFail {
+                        string: format!("{}", x),
+                    };
+                    if let Ok(int) = x.parse::<i64>() {
+                        return Ok(TensorAxisIndex::Single(int));
+                    }
+                    if let Some(no_t) = x.strip_prefix("t") {
+                        return tensor_from_hash_prefix(no_t)
+                            .map(|z| TensorAxisIndex::Tensor(z))
+                            .map_err(|_e| CircuitConstructionError::TensorHashNotFound {
+                                hash: no_t.to_owned(),
+                            });
+                    }
+                    if let [start, end] = x.split(":").collect::<Vec<_>>()[..] {
+                        return Ok(TensorAxisIndex::Slice(Slice {
+                            start: if start.is_empty() {
+                                None
+                            } else {
+                                Some(start.parse::<i64>().map_err(|_e| num_er.clone())?)
+                            },
+                            stop: if end.is_empty() {
+                                None
+                            } else {
+                                Some(end.parse::<i64>().map_err(|_e| num_er.clone())?)
+                            },
+                        }));
+                    }
+                    return Err(er.clone());
+                })
+                .collect::<Result<Vec<_>, CircuitConstructionError>>()?,
+        ))
+    }
 }
 
 impl IntoPy<PyObject> for TensorIndex {
